import sys
sys.path.append(r"c:\\users\\erdoel\\documents\\works\\python_works\\microwave_toolbox")
from mwtoolbox.rfnetwork import *
import network
import itertools
import numpy as np
frequencies=np.linspace(57e9,63.5e9,8,endpoint=True)
print(frequencies)
sptline=spfile(noktasayisi=len(frequencies),portsayisi=2)
theta=90
sptline.set_sparam_gen_func(lambda x: network.abcd2s(network.tline(100.0,x*theta/20e9*np.pi/180.0),50.0))
sptline.set_frequency_points(frequencies)

import matplotlib.pyplot as plt
plt.plot(frequencies,sptline.S(1,1,"DB"))
plt.show()
# sptline.set_sparam_gen_func(None)
frequencies=np.linspace(1e9,63.5e9,80,endpoint=True)
sptline.set_frequency_points(frequencies)
import matplotlib.pyplot as plt
plt.plot(frequencies,sptline.S(1,1,"DB"))
plt.show()



#-*-coding:utf-8-*-
from .genel import *
speed_of_light_in_freespace=299792458*pq.m/pq.s
free_space_wave_impedance=376.730313461*pq.ohm
free_space_permeability=12.566370614e-7*pq.N/pq.A/pq.A
free_space_permittivity=8.854187817e-12*pq.F/pq.m

mu0 = free_space_permeability.simplified.magnitude
eps0 = free_space_permittivity.simplified.magnitude
co = c0 = speed_of_light_in_freespace.simplified.magnitude



#-*-coding:utf-8-*-
"""
General methods
"""
import sys
import quantities as pq
import numpy as np
import re
import collections
import cProfile
from copy import deepcopy
import os
from functools import lru_cache

Np2dB = 8.68589 # coefficient to convert from alpha to dB/m (20.0*log10(e))
                # unit of alpha is 1/m (Neper)
globsep = ";"
globsep2 = ":"  # empedans degerlerini ayirmak icin
### New Units For Quantities Library
pq.mH = pq.millihenry = pq.UnitQuantity('millihenry', pq.H / 1000, symbol='mH')
pq.uH = pq.microhenry = pq.UnitQuantity('microhenry', pq.H / 1e6, symbol='uH')
pq.nH = pq.nanohenry = pq.UnitQuantity('nanohenry', pq.H / 1e9, symbol='nH')
pq.pH = pq.picohenry = pq.UnitQuantity('picohenry', pq.H / 1e12, symbol='pH')

pq.mF = pq.millifarad = pq.UnitQuantity('millifarad', pq.F / 1000, symbol='mF')
pq.uF = pq.microfarad = pq.UnitQuantity('microfarad', pq.F / 1e6, symbol='uF')
pq.nF = pq.nanofarad = pq.UnitQuantity('nanofarad', pq.F / 1e9, symbol='nF')
pq.pF = pq.picofarad = pq.UnitQuantity('picofarad', pq.F / 1e12, symbol='pF')

pq.microinch = pq.UnitQuantity('microinch', pq.inch / 1e6, symbol='microinch')

pq.dB = pq.UnitQuantity('dB', pq.dimensionless, symbol='dB')

def ekpolyfit(x):
    """Polynomial fit for ellipk function.
    Works for argument from 0 to 0.98 with good accuracy.
    """
    coefs = [1.571, 0.3972, 0.09132, 1.807, -11.16, 44.75, -106, 148.9, -114, 36.94]
    return np.sum([coefs[i] * x ** i for i in range(len(coefs))])

def tukey_window(alpha,N):
    """
    Tukey window (also known as "tapered cosine window")
    Also available in scipy.signal
    """
    sonuc=[]
    cos = np.cos
    pi = np.pi
    for i in range(N):
        if (i<=alpha*(N-1.0)/2.0):
            sonuc.append(0.5*(1.0+cos(pi*(2.0*i/alpha/(N-1.0)-1.0))))
        elif (i<=(N-1)*(1.0-alpha/2.0)):
            sonuc.append(1.0)
        elif (i<=(N-1)):
            sonuc.append(0.5*(1.0+cos(pi*(2.0*i/alpha/(N-1.0)-2.0/alpha+1.0))))
    return sonuc

def blackman_window(N):
    """
    Blackman-Harris window
    Also available in scipy.signal
    """
    a0 = 0.35875
    a1 = 0.48829
    a2 = 0.14128
    a3 = 0.01168
    arr = np.linspace(0,N-1,N)
    return a0-a1*np.cos(2*np.pi*arr/(N-1))+a2*np.cos(4*np.pi*arr/(N-1))-a3*np.cos(6*np.pi*arr/(N-1))

def gaussian_window(sigma,N):
    """
    Gaussian window
    sigma should be smaller than or equal to 0.5
    Also available in scipy.signal
    Ref: Wikipedia
    """
    arr = np.linspace(0,N-1,N)
    return np.exp(-0.5*((arr-(N-1.0)/2.0)/(sigma*(N-1.0)/2.0))**2)

def cmp(x, y):
    """
    Replacement for built-in function cmp that was removed in Python 3

    Compare the two objects x and y and return an integer according to
    the outcome. The return value is negative if x < y, zero if x == y
    and strictly positive if x > y.
    """
    return int(x > y) - int(x < y)

try:
    basestring
except NameError:
    basestring = str

try:
    unicode
except NameError:
    unicode = bytes


def prettystring(miktarlar, birim=None):
    formatstring1 = '{:3.3e}'
    formatstring2 = '{:3.3f}'
    if len(np.shape(miktarlar)) == 0:
        miktarlar = [miktarlar]
    liste = []
    for i in range(len(miktarlar)):
        miktar = miktarlar[i]
        if (np.fabs(float(miktar))==np.inf):
            liste.append(str(miktar))
            continue
        if isinstance(miktar, pq.Quantity):
            if birim == "SI":
                miktar = miktar.simplified
                dims = miktar.dimensionality.string
                miktar = miktar.magnitude
            elif birim == "dBm":
                temp = miktar.simplified.magnitude
                miktar = 10.0 * np.log10(temp * 1000.0)
                dims = birim
            elif birim is not None:
                miktar.units = birim
                miktar = miktar.magnitude
                dims = birim
            else:
                miktar = miktar.magnitude
                dims = miktar.dimensionality.string
            if dims == "dimensionless":
                dims = ""
            miktar = float(miktar)
            if miktar<0.001:
                formatstring = formatstring1
            else:
                formatstring = formatstring2
            liste.append((formatstring.format(miktar) + " " + dims).strip())
        elif isinstance(miktar, (float, int, np.ndarray)):
            miktar = float(miktar)
            if birim == "dBm":
                miktar = 10.0 * np.log10(miktar * 1000.0)
                dims = birim
            elif birim is not None:
                miktar = pq.Quantity(miktar * coef(birim), birim).magnitude
                dims = birim
            else:
                dims = ""
            if dims == "dimensionless":
                dims = ""
            miktar = float(miktar)
            if miktar<0.001:
                formatstring = formatstring1
            else:
                formatstring = formatstring2
            liste.append((formatstring.format(miktar) + " " + dims).strip())
        elif isinstance(miktar, str):
            temp = convert2pq(miktar)[0]
            if birim is not None:
                temp = pq.Quantity(temp * coef(birim), birim)
                dims = birim
                if float(temp.magnitude)<0.001:
                    formatstring = formatstring1
                else:
                    formatstring = formatstring2
                liste.append((formatstring.format(float(temp.magnitude)) + " " + dims).strip())
            else:
                liste.append(miktar)
    return globsep.join(liste)

def coef(outputunit):
    """ Coefficient to convert from SI unit to outputunit """
    temp = pq.Quantity(1.0, outputunit)
    return (temp.magnitude) / (temp.simplified.magnitude)

# removed from this module, use sep_unit(sayi)[0] instead
# def stripunit(sayi):
#     match = re.search(r"([+\-]?)(\d+(\.\d*)?|\d*\.\d+)([eE][+\-]?\d+)?\s*(\D+\S*)?",sayi)
#     number = ""
#     unit = ""
#     if match is not None:
#         for k in [1, 2, 4]:
#             if match.group(k) is not None:
#                 number = number + str(match.group(k))
#     return float(number)

def sep_unit(sayi):
    """This function parses input string representing a number with unit and returns number and unit separately"""
    match = re.search(r"([+\-]?)(\d+(\.\d*)?|\d*\.\d+)([eE][+\-]?\d+)?\s*(\D+\S*)?",sayi)
    number = ""
    unit = ""
    if match is not None:
        for k in [1, 2, 4]:
            if match.group(k) is not None:
                number = number + str(match.group(k))
    return (float(number), match.group(5))

# @lru_cache did not work with arguments of type list since lists are not hashable
def convert2pq(sayilar, defaultunits=None):
    """
    Method to convert a string or string list to float after unit conversion to SI
    Units are extracted from strings.
    If there is not a unit in string, unit is taken from defaultunits
    """

    number = ""
    # unit = ""
    if isinstance(sayilar, (float,int)):
        return [float(sayilar)]
    if isinstance(sayilar, str):
        sayilar = [sayilar]
    if not defaultunits:
        defaultunits=[""]*len(sayilar)
    if isinstance(defaultunits, str):
        defaultunits = [defaultunits]
    sonuc = []

    if not hasattr(convert2pq, "sayilar"):   # make sayilar static variable
        convert2pq.sayilar = deepcopy(sayilar)
        convert2pq.defaultunits=deepcopy(defaultunits)
    elif (convert2pq.sayilar == sayilar) and convert2pq.defaultunits == defaultunits:
        return convert2pq.sonuc

#    if not hasattr(convert2pq, "regex"):
         # make regex static variable, did not provide any performance advantage
#        # pattern for matching real numbers
#        pattern = r"([+\-]?)(\d+(\.\d*)?|\d*\.\d+)([eE][+\-]?\d+)?\s*(\D+\S*)?"
#        convert2pq.regex = re.compile(pattern)

    sonuclar = []
    for i in range(len(sayilar)):
        sayi = sayilar[i]
        try:
            if (sayi == convert2pq.sayilar[i]) and (convert2pq.units[i]):
                sonuclar.append(convert2pq.sonuc[i])
                continue
        except:
            pass
        if isinstance(sayi, pq.Quantity):
            sonuc = sayi
            sonuclar.append(float(sonuc.simplified.magnitude))
        elif isinstance(sayi, (float, np.ndarray, int, list)):
            # unit = ""
            # if len(defaultunits) > 0:
                # unit = defaultunits[i]
            # sonuc=pq.Quantity(float(sayi), unit)
            sonuc=pq.Quantity(np.array(sayi), defaultunits[i])
            sonuclar.append(sonuc.simplified.magnitude)
        else:
            sayi = sayi.split("=")[-1].strip()
            #match = convert2pq.regex.search(sayi)
            match = re.search(r"([+\-]?)(\d+(\.\d*)?|\d*\.\d+)([eE][+\-]?\d+)?\s*(\D+\S*)?",sayi)
            number = ""
            unit = ""
            if match is not None:
                for k in [1, 2, 4]:
                    if match.group(k) is not None:
                        number = number + str(match.group(k))
                unit = str(match.group(5))
                if unit == "None":
                    # unit = ""
                    # if len(defaultunits) > 0:
                        # unit = defaultunits[i]
                    unit = defaultunits[i]
                sonuc = pq.Quantity(float(number), unit)
            else:
                sonuc = None
            if sonuc:
                sonuclar.append(float(sonuc.simplified.magnitude))
            else: # sonuc = None
                print("None input is detected and assumed to be 0!")
                sonuclar.append(0.0)

    convert2pq.sonuc = deepcopy(sonuclar)
    convert2pq.sayilar = deepcopy(sayilar)
    convert2pq.defaultunits=deepcopy(defaultunits)
    return convert2pq.sonuc

def flatten(x):
    """Flatten (an irregular) list of lists"""
    result = []
    for el in x:
        if hasattr(el, "__iter__") and not isinstance(el, basestring):
            result.extend(flatten(el))
        else:
            result.append(el)
    return result

def change_format(data, data_format="complex"):
    """Changes format of complex data to other formats."""
    data_format = data_format.lower()
    if data_format == "complex":
        return data
    elif data_format == "real":
        return np.real(data)
    elif data_format == "imag":
        return np.imag(data)
    elif data_format == "mag":
        return np.abs(data)
    elif data_format == "phase":
        return np.angle(data, deg=True)
    elif data_format == "uphase":
        return np.unwrap(np.angle(data)) * 180.0 / np.pi

def heatmap(data, row_labels, col_labels, ax=None, cbar_kw={}, cbarlabel="", **kwargs):
    """
    Create a heatmap from a numpy array and two lists of labels.
    https://matplotlib.org/3.5.0/gallery/images_contours_and_fields/image_annotated_heatmap.html

    Args:
        data: A 2D numpy array of shape (M, N).
        row_labels: A list or array of length M with the labels for the rows.
        col_labels: A list or array of length N with the labels for the columns.
        ax: A `matplotlib.axes.Axes` instance to which the heatmap is plotted.  If not provided, use current axes or create a new one.  Optional.
        cbar_kw: A dictionary with arguments to `matplotlib.Figure.colorbar`.  Optional.
        cbarlabel: The label for the colorbar.  Optional.
        kwargs: All other arguments are forwarded to `imshow`.

    Usage:
        fig, ax = plt.subplots()
        im, cbar = heatmap(harvest, vegetables, farmers, ax=ax, cmap="YlGn", cbarlabel="harvest [t/year]")
        texts = annotate_heatmap(im, valfmt="{x:.1f} t")

    """
    import matplotlib.pyplot as plt
    if not ax:
        ax = plt.gca()

    # Plot the heatmap
    im = ax.imshow(data, **kwargs)

    # Create colorbar
    cbar = ax.figure.colorbar(im, ax=ax, **cbar_kw)
    cbar.ax.set_ylabel(cbarlabel, rotation=-90, va="bottom")

    # Show all ticks and label them with the respective list entries.
    ax.set_xticks(np.arange(data.shape[1]), labels=col_labels)
    ax.set_yticks(np.arange(data.shape[0]), labels=row_labels)

    # Let the horizontal axes labeling appear on top.
    ax.tick_params(top=True, bottom=False,
                   labeltop=True, labelbottom=False)

    # Rotate the tick labels and set their alignment.
    plt.setp(ax.get_xticklabels(), rotation=-30, ha="right",
             rotation_mode="anchor")

    # Turn spines off and create white grid.
    ax.spines[:].set_visible(False)

    ax.set_xticks(np.arange(data.shape[1]+1)-.5, minor=True)
    ax.set_yticks(np.arange(data.shape[0]+1)-.5, minor=True)
    ax.grid(which="minor", color="w", linestyle='-', linewidth=3)
    ax.tick_params(which="minor", bottom=False, left=False)

    return im, cbar


def annotate_heatmap(im, data=None, valfmt="{x:.2f}", textcolors=("black", "white"), threshold=None, **textkw):
    """
    A function to annotate a heatmap.
    https://matplotlib.org/3.5.0/gallery/images_contours_and_fields/image_annotated_heatmap.html

    Args:
    	im: The AxesImage to be labeled.
    	data: Data used to annotate.  If None, the image's data is used.  Optional.
    	valfmt: The format of the annotations inside the heatmap.  This should either use the string format method, e.g. "$ {x:.2f}", or be a `matplotlib.ticker.Formatter`.  Optional.
    	textcolors: A pair of colors.  The first is used for values below a threshold, the second for those above.  Optional.
    	threshold: Value in data units according to which the colors from textcolors are applied.  If None (the default) uses the middle of the colormap as separation.  Optional.
    	kwargs: All other arguments are forwarded to each call to `text` used to create the text labels.

    Usage:
        fig, ax = plt.subplots()
        im, cbar = heatmap(harvest, vegetables, farmers, ax=ax, cmap="YlGn", cbarlabel="harvest [t/year]")
        texts = annotate_heatmap(im, valfmt="{x:.1f} t")

    """
    import matplotlib

    if not isinstance(data, (list, np.ndarray)):
        data = im.get_array()

    # Normalize the threshold to the images color range.
    if threshold is not None:
        threshold = im.norm(threshold)
    else:
        threshold = im.norm(data.max())/2.

    # Set default alignment to center, but allow it to be
    # overwritten by textkw.
    kw = dict(horizontalalignment="center",
              verticalalignment="center")
    kw.update(textkw)

    # Get the formatter in case a string is supplied
    if isinstance(valfmt, str):
        valfmt = matplotlib.ticker.StrMethodFormatter(valfmt)

    # Loop over the data and create a `Text` for each "pixel".
    # Change the text's color depending on the data.
    texts = []
    for i in range(data.shape[0]):
        for j in range(data.shape[1]):
            kw.update(color=textcolors[int(im.norm(data[i, j]) > threshold)])
            text = im.axes.text(j, i, valfmt(data[i, j], None), **kw)
            texts.append(text)

    return texts

def smooth(x, window_len=11, window='hanning'):
    """smooth the data using a window with requested size.
    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal
    (with the window size) in both ends so that transient parts are minimized
    in the begining and end part of the output signal.

    Args:
        x(ndarray): the input signal
        window_len(int, optional): the dimension of the smoothing window; should be an odd integer
        window((string, list, ndarray), optional): either
                    window array with type list or numpy array with size window_len
                or
                    the type of window from 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'

    Returns:
        the smoothed signal

    Example:
        t=linspace(-2,2,0.1)
        x=sin(t)+randn(len(t))*0.1
        y=smooth(x)
        see also:
        numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman, numpy.convolve
        scipy.signal.lfilter

    """

    import numpy as np

    if x.ndim != 1:
        raise ValueError("smooth only accepts 1 dimension arrays.")
    if x.size < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")
    if window_len < 3:
        return x
    if not window in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
        raise ValueError("Window is on of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'")
    s = np.r_[2 * x[0] - x[window_len:1:-1], x, 2 * x[-1] - x[-1:-window_len:-1]]
    if isinstance(window, (list,np.ndarray)):
        w = np.array(window)
    else:
        if window == 'flat':  #moving average
            w = np.ones(window_len, 'd')
        else:
            w = eval('np.' + window + '(window_len)')
    y = np.convolve(w / w.sum(), s, mode='same')
    return y[window_len - 1:-window_len + 1]

def str_distance(s, t):
    """ levenshtein_ratio_and_distance:
        Calculates levenshtein distance between two strings.
        If ratio_calc = True, the function computes the
        levenshtein distance ratio of similarity between two strings
        For all i and j, distance[i,j] will contain the Levenshtein
        distance between the first i characters of s and the
        first j characters of t
    """
    # Initialize matrix of zeros
    rows = len(s)+1
    cols = len(t)+1
    distance = np.zeros((rows,cols),dtype = int)

    # Populate matrix of zeros with the indeces of each character of both strings
    for i in range(1, rows):
        for k in range(1,cols):
            distance[i][0] = i
            distance[0][k] = k

    # Iterate over the matrix to compute the cost of deletions,insertions and/or substitutions
    for col in range(1, cols):
        for row in range(1, rows):
            if s[row-1] == t[col-1]:
                cost = 0 # If the characters are the same in the two strings in a given position [i,j] then the cost is 0
            else:
                # In order to align the results with those of the Python Levenshtein package, if we choose to calculate the ratio
                # the cost of a substitution is 2. If we calculate just distance, then the cost of a substitution is 1.
                cost = 1
            distance[row][col] = min(distance[row-1][col] + 1,      # Cost of deletions
                                 distance[row][col-1] + 1,          # Cost of insertions
                                 distance[row-1][col-1] + cost)     # Cost of substitutions

    # Computation of the Levenshtein Distance Ratio
    Ratio = ((len(s)+len(t)) - distance[row][col]) / (len(s)+len(t))
    # print(distance) # Uncomment if you want to see the matrix showing how the algorithm computes the cost of deletions,
    # insertions and/or substitutions
    # This is the minimum number of edits needed to convert string a to string b
    # return "The strings are {} edits away".format(distance[rows-1][cols-1])
    return distance[rows-1][cols-1], Ratio


def crossing(xvals, yvals, level):
    """
    This function calculates the x-values at which yvals array crosses level value. Linear interpolation is used.
    
    Args:
        xvals(np.ndarray): x-values at which values of input array is given.
        yvals(np.ndarray): y-values corresponding to x-values.
        level(float): Crossing value

    Return:
        np.ndarray: x-values at crossing points.
    """
    arr1 = yvals - level
    sarr1 = np.sign(arr1)
    indices = np.array([i for i in range(len(sarr1)-1) if (sarr1[i+1]+sarr1[i])==0 or sarr1[i+1]==0])
    print(f"{indices=}")
    if len(indices)==0:
        return None
    else:
        f1 = xvals[indices  ]
        f2 = xvals[indices+1]
        y1 = np.abs(arr1[indices  ])
        y2 = np.abs(arr1[indices+1])
        return (y1 * f2 + y2 * f1) / (y1 + y2)


if __name__ == "__main__":
      #rr = polarsample(0.2)
    #print(rr)
    #for r in rr:
    #    print(r)
    #import matplotlib.pyplot as plt
    #import pysmith
    #fig= plt.figure(figsize=(15,10))
    #ax = pysmith.get_smith(fig, 111)
    #for r in rr:
    #    ax.plot([np.real(r)],[np.imag(r)],"*")
    #plt.savefig("Sample_points_02.png")
    #plt.show()

   # arg = ["12.1", "34.2", "23.4"]
   # arg = [np.array([12.1, 15.2, 18.3]), "34.2", "23.4"]
   # arg1 = Tee_Attenuator_Analysis(arg, ["ohm", "ohm", "ohm", "", "", "", "", ""])
    # import time
    # time.clock()
    # for i in np.linspace(10,50,10000):
        # # convert2pq([i,"20","1","0","0","0","0","0","0","0","0"],[])
        # convert2pq([i,"20"],[])
    # print time.clock()
    # for i in range(10,50,1000):
        # convert2pq_eski([10,"20","1","0","0","0","0","0","0","0","0"],[])

    # print convert2pq([10.0], defaultunits=["GHz"])
    # print convert2pq([10.0], defaultunits=["GHz"])
    # print convert2pq([10.0], defaultunits=["Hz"])
    # print convert2pq([10.0], defaultunits=["MHz"])
    # print convert2pq([100.0], defaultunits=["MHz"])

    # print(str_distance("rx1","rx2", ratio_calc = False))
    dat = blackman_window(100)
    print(dat)
    pass



# TODO: add annotaiton like:
# ax2.annotate(f"{fhigh_hl}GHz", xy=(np.real(data[-1]), np.imag(data[-1])), xytext=(-100,0), textcoords="offset pixels", arrowprops=dict(facecolor='black', shrink=0.05))
import numpy as np
import matplotlib
import matplotlib.pyplot as plt

def get_smith(fig, rect=111, plot_impedance=True, plot_ticks=False, plot_admittance=True, plot_labels=True,
              resticks=None, reactticks=None, xlim=(-1,1), ylim=(-1,1)):

    # font definition
    font = {'family': 'sans-serif',
            'color':  'black',
            'weight': 'normal',
            'size': 18,
            }
    if isinstance(rect, int):
        ax = fig.add_subplot(rect)
    else: #3-tuple
        ax = fig.add_subplot(*rect)
    ax.spines[["bottom"]].set_position("center")
    ax.spines[["left","top","right"]].set_visible(False)
    ax.set_xlim(list(xlim))
    ax.set_ylim(list(ylim))
    ax.set_yticks([])
    ax.spines['bottom'].set_position('zero')

    if plot_impedance:
    
        # make lines of constant resistance
        if resticks is None:
            res_log1 = np.linspace(-4,4,9)
            res1 = 2**res_log1
        else:
            res1 = np.asarray(resticks[:])
        react_log1 = np.linspace(-5,5,2001)
        react = 10**react_log1
        react2 = np.append(-1.0*react[::-1],np.array([0]))
        react = np.append(react2,react)
        for r in res1:
            z = 1j*react + r
            gam = (z-1)/(z+1)
            x = np.real(gam)
            y = np.imag(gam)
            if abs(r-1) > 1e-6:
                ax.plot(x,y,'-k',linewidth = 0.5,alpha=0.5)
            else:
                ax.plot(x,y,'-k',linewidth = 1.0,alpha=1.0)

        # make lines of constant reactance
        if reactticks is None:
            react_log1 = np.linspace(-3,3,7)
            react = 2**react_log1
        else:
            react = np.asarray(reactticks[:])
        res_log1 = np.linspace(-5,5,2001)
        res = 10**res_log1
        react = np.append(-1.0*react[::-1],react)
        for chi in react:
            z = 1j*chi + res
            gam = (z-1)/(z+1)
            x = np.real(gam)
            y = np.imag(gam)
            if abs(chi-1) > 1e-6 and abs(chi+1) > 1e-6 and abs(chi) > 1e-6:
                ax.plot(x,y,'-k',linewidth = 0.5,alpha=0.5)
            else:
                ax.plot(x,y,'-k',linewidth = 1.0,alpha=1.0)
            if min(x)>xlim[1] or max(x)<xlim[0] or min(y)>ylim[1] or max(y)<ylim[0]:
                continue
            x1, x2 = max(xlim[0], x[0]), min(xlim[1], x[-1])
            y1, y2 = y[np.argmin(np.abs(x-x1))], y[np.argmin(np.abs(x-x2))]
            y1, y2 = max(ylim[0], min(y1,y2)), min(ylim[1], max(y1,y2))
            x1, x2 = x[np.argmin(np.abs(y-y1))], x[np.argmin(np.abs(y-y2))]
            xo, yo = (x1, y1) if x1<x2 else (x2, y2)
            if plot_labels:
                text = ax.text(xo, yo, f"{chi:.02f}", fontsize=font["size"]/3)

    if plot_admittance:

        # make lines of constant conductance
        if resticks is None:
            res_log1 = np.linspace(-4,4,9)
            res1 = 2**res_log1
        else:
            res1 = np.asarray(resticks[:])
        react_log1 = np.linspace(-5,5,2001)
        react = 10**react_log1
        react = np.append(-1.0*react[::-1],react)
        for r in res1:
            y = 1.0/r + 1.0/(1j*react)
            gam = (1.0/y-1)/(1.0/y+1)
            x = np.real(gam)
            y = np.imag(gam)
            if abs(r-1) > 1e-6:
                ax.plot(x,y,'-k',linewidth = 0.5,alpha=0.5)
            else:
                ax.plot(x,y,'-k',linewidth = 1.0,alpha=1.0)

        # make lines of constant susceptance
        if reactticks is None:
            react_log1 = np.linspace(-3,3,7)
            react = 2**react_log1
        else:
            react = np.asarray(reactticks[:])
        res_log1 = np.linspace(-5,5,2001)
        res = 10**res_log1
        react = np.append(-1.0*react[::-1],react)
        for chi in react:
            y = 1.0/(1j*chi) + 1.0/res
            gam = (1.0/y-1)/(1.0/y+1)
            x = np.real(gam)
            y = np.imag(gam)
            if abs(chi-1) > 1e-6 and abs(chi+1) > 1e-6 and abs(chi) > 1e-6:
                ax.plot(x,y,'-k',linewidth = 0.5,alpha=0.5)
            else:
                ax.plot(x,y,'-k',linewidth = 1.0,alpha=1.0)
            if min(x)>xlim[1] or max(x)<xlim[0] or min(y)>ylim[1] or max(y)<ylim[0]:
                continue
            x1, x2 = max(xlim[0], x[0]), min(xlim[1], x[-1])
            y1, y2 = y[np.argmin(np.abs(x-x1))], y[np.argmin(np.abs(x-x2))]
            y1, y2 = max(ylim[0], min(y1,y2)), min(ylim[1], max(y1,y2))
            x1, x2 = x[np.argmin(np.abs(y-y1))], x[np.argmin(np.abs(y-y2))]
            xo, yo = (x1, y1) if x1>x2 else (x2, y2)
            if plot_labels:
                text = ax.text(xo, yo, f"{chi:.02f}", fontsize=font["size"]/3)


    theta = np.linspace(-np.pi, np.pi, 10000)
    x = np.cos(theta)
    y = np.sin(theta)
    ax.plot(x,y,'-k',linewidth=0.5, alpha=0.5)

    if plot_labels:
        xt = np.array([xa for xa in (res1-1)/(res1+1) if xlim[1]>xa>xlim[0]])
        res1 = (1+xt)/(1-xt)
        ax.set_xticks(xt)
        ax.set_xticklabels([f"{rr:.2f}" if rr<1000 else "$\infty$" for rr in res1], fontsize = font["size"]/3)

    ax.set_aspect("equal")
    return ax


if __name__ == "__main__":
    
    fig = plt.figure(figsize=(5,5))
    ax = get_smith(fig, 111, plot_ticks=True, plot_admittance=False, plot_labels=True)
    plt.show()



from mwtoolbox import smithplot
import matplotlib.pyplot as plt 
import numpy as np

# fig1 = plt.figure(figsize=(15, 7), layout="constrained")
# ax1 = smithplot.get_smith(fig1, 121, plot_impedance=True, plot_ticks=False, plot_admittance=True, plot_labels=True,)

fig2 = plt.figure(figsize=(15, 7), layout="constrained")
ax2 = smithplot.get_smith(fig2, 121, plot_impedance=True,
                          plot_ticks=False, plot_admittance=True, plot_labels=True,
                          # resticks=[0.5,1.0,1.5,2.0],
                          xlim=(-1,0.5), ylim=(-1,0.5))

# fig3 = plt.figure(figsize=(15, 7), layout="constrained")
# ax3 = smithplot.get_smith(fig3, 121, plot_impedance=False, plot_ticks=False, plot_admittance=True, plot_labels=True,)

# for ax in [ax2]:
#     ax.set_xlim([-0.5,0.5])
#     ax.set_ylim([-0.5,0.5])

# fig1.savefig("only_admittance_impedance.png")
# fig2.savefig("only_impedance.png")
# fig3.savefig("only_admittance.png")

plt.show()



# -*- coding:utf-8 -*-
# Mode-Matching code to analyse waveguide junction
# Reference: Computer Aided Design of Waveguide Devices By Mode-Matching Method
import numpy as np
from numpy.lib.scimath import sqrt as csqrt
from scipy.integrate import dblquad
import itertools
import time
current=time.time()

mu0=4*np.pi*1e-7
eps0=8.854187817e-12
c0=299792458
pi = np.pi

def isqrt(k2):
    if k2>0:
        kz = np.sqrt(k2)
    else:
        kz= -1j*np.sqrt(-k2)
    return kz

def Ex(x,y,xc,yc,m,n,a,b,eps,freq,mode=1):
    """ Mode=1 for TE, 0 for TM """
    w = 2*pi*freq
    kc = np.sqrt((n*pi/b)**2+(m*pi/a)**2)
    k2 = w**2*eps/c0**2-kc**2
    kz = isqrt(k2)
    if mode:
        return 1j*w*mu0/kc**2*(n*pi/b)*np.cos(m*pi*(x-xc)/a)*np.sin(n*pi*(y-yc)/b)
    else:
        return -1j*kz/kc**2*(m*pi/a)*np.cos(m*pi*(x-xc)/a)*np.sin(n*pi*(y-yc)/b)

def Ey(x,y,xc,yc,m,n,a,b,eps,freq,mode=1):
    """ Mode=1 for TE, 0 for TM """
    w  = 2*pi*freq
    kc = np.sqrt((n*pi/b)**2+(m*pi/a)**2)
    k2 = w**2*eps/c0**2-kc**2
    kz = isqrt(k2)
    if mode:
        return -1j*w*mu0/kc**2*(m*pi/a)*np.sin(m*pi*(x-xc)/a)*np.cos(n*pi*(y-yc)/b)
    else:
        return -1j*kz/kc**2*(n*pi/b)*np.sin(m*pi*(x-xc)/a)*np.cos(n*pi*(y-yc)/b)

def Hx(x,y,xc,yc,m,n,a,b,eps,freq,mode=1):
    """ Mode=1 for TE, 0 for TM """
    w  = 2*pi*freq
    kc = np.sqrt((n*pi/b)**2+(m*pi/a)**2)
    k2 = w**2*eps/c0**2-kc**2
    kz = isqrt(k2)
    if mode:
        return 1j*kz/kc**2*(m*pi/a)*np.sin(m*pi*(x-xc)/a)*np.cos(n*pi*(y-yc)/b)
    else:
        return 1j*w*eps0*eps/kc**2*(n*pi/b)*np.sin(m*pi*(x-xc)/a)*np.cos(n*pi*(y-yc)/b)

def Hy(x,y,xc,yc,m,n,a,b,eps,freq,mode=1):
    """ Mode=1 for TE, 0 for TM """
    w  = 2*pi*freq
    kc = np.sqrt((n*pi/b)**2+(m*pi/a)**2)
    k2 = w**2*eps/c0**2-kc**2
    kz = isqrt(k2)
    if mode:
        return 1j*kz/kc**2*(n*pi/b)*np.cos(m*pi*(x-xc)/a)*np.sin(n*pi*(y-yc)/b)
    else:
        return -1j*w*eps*eps0/kc**2*(m*pi/a)*np.cos(m*pi*(x-xc)/a)*np.sin(n*pi*(y-yc)/b)

def eh(x,y,params):
    return Ex(x,y,*params)*Hy(x,y,*params)-Ey(x,y,*params)*Hx(x,y,*params)

def createmesh(xcs,ycs,xcf,ycf,Nx,Ny):
    dx=(xcf-xcs)/Nx
    dy=(ycf-ycs)/Ny
    xp = np.linspace(xcs+dx/2,xcf-dx/2,Nx)
    yp = np.linspace(ycs+dy/2,ycf-dy/2,Ny)
    xp_1, yp_1 = np.meshgrid(xp,yp)
    dS = dx*dy
    return xp_1, yp_1, dS

def exh(x,y,par1,par2):
    return Ex(x,y,*par2)*Hy(x,y,*par1)-Ey(x,y,*par2)*Hx(x,y,*par1)

def sp_wg(a,b,eps,L,modes):
    temp=[np.exp(-L*1j*isqrt(w**2*eps/c0**2-((n*pi/b)**2+(m*pi/a)**2))) for m,n,k in modes]
    return np.matrix(np.diag(temp))

class MM_RWG:
    def __init__(self,wg1, wg2, aperture=None):
        self.xc1=wg1[0]
        self.yc1=wg1[1]
        self.xc2=wg2[0]
        self.yc2=wg2[1]
        self.a1=wg1[2]
        self.b1=wg1[3]
        self.a2=wg2[2]
        self.b2=wg2[3]
        self.eps1=wg1[4]
        self.eps2=wg2[4]
        self.aperture = aperture

    def simulate(self, freqs):
        xc1 = self.xc1
        yc1 = self.yc1
        xc2 = self.xc2
        yc2 = self.yc2
        a1 = self.a1
        b1 = self.b1
        a2 = self.a2
        b2 = self.b2
        eps1 = self.eps1
        eps2 = self.eps2
        Nx=40
        Ny=20
        freq_cutoff=np.max(freqs)*5
        # Boundaries of aperture
        if self.aperture:
            xt2 = self.aperture[1]
            xt1 = self.aperture[0]
            yt2 = self.aperture[3]
            yt1 = self.aperture[2]
        else:
            xt2 = np.min([xc1+a1,xc2+a2])
            xt1 = np.max([xc1,xc2])
            yt2 = np.min([yc1+b1,yc2+b2])
            yt1 = np.max([yc1,yc2])
        print(xt1,xt2,yt1,yt2)
        # Maximum mode numbers that are taken into account
        n1=int(2*pi*freq_cutoff/c0*b1/pi)
        m1=int(2*pi*freq_cutoff/c0*a1/pi)
        n2=int(2*pi*freq_cutoff/c0*b2/pi)
        m2=int(2*pi*freq_cutoff/c0*a2/pi)
        n3=int(2*pi*freq_cutoff/c0*(yt2-yt1)/pi)
        m3=int(2*pi*freq_cutoff/c0*(xt2-xt1)/pi)

        # These are list of modes, each element is a 3-tuple (m,n,k)
        #   m: wavenumber in x-axis
        #   n: wavenumber in y-axis
        #   k: type of mode (1:TE, 0:TM)
        # modes1: Mode expansion at WG-1
        # modes2: Mode expansion at WG-2
        # modes3: Mode expansion at aperture
        TEmodes1= list(itertools.product(list(range(m1+1)),list(range(n1+1)),[1]))
        TEmodes2= list(itertools.product(list(range(m2+1)),list(range(n2+1)),[1]))
        TEmodes3= list(itertools.product(list(range(m3+1)),list(range(n3+1)),[1]))
        TMmodes1= list(itertools.product(list(range(1,m1+1)),list(range(1,n1+1)),[0]))
        TMmodes2= list(itertools.product(list(range(1,m2+1)),list(range(1,n2+1)),[0]))
        TMmodes3= list(itertools.product(list(range(1,m3+1)),list(range(1,n3+1)),[0]))
        TEmodes1.pop(TEmodes1.index((0,0,1))) # remove undefined modes
        TEmodes2.pop(TEmodes2.index((0,0,1))) # remove undefined modes
        TEmodes3.pop(TEmodes3.index((0,0,1))) # remove undefined modes
        modes1 = TEmodes1 + TMmodes1
        modes2 = TEmodes2 + TMmodes2
        modes3 = TEmodes3 + TMmodes3
        modes13= list(itertools.product(modes1, modes3))
        modes23= list(itertools.product(modes2, modes3))
        Nw1 = len(modes1)
        Nw2 = len(modes2)
        Ns  = len(modes3)

        xp_1, yp_1, dS1 = createmesh(xc1,yc1,xc1+a1,yc1+b1,Nx,Ny)
        xp_2, yp_2, dS2 = createmesh(xc2,yc2,xc2+a2,yc2+b2,Nx,Ny)
        xp_3, yp_3, dS3 = createmesh(xt1,yt1,xt2,yt2,Nx,Ny)

        Iw1 = np.matrix(np.eye(Nw1))
        Iw2 = np.matrix(np.eye(Nw2))
        Is  = np.matrix(np.eye(Ns))
        SP  = np.matrix(np.zeros((Nw1+Nw2,Nw1+Nw2),dtype=complex))
        Qw1 = np.matrix(np.zeros((Nw1,Nw1),dtype=complex))
        Qw2 = np.matrix(np.zeros((Nw2,Nw2),dtype=complex))
        Qs  = np.matrix(np.zeros((Ns,Ns),dtype=complex))

        SPARAMS = []
        for freq in freqs:
            print(freq)
            for i in range(Nw1):
                m,n,mode = modes1[i]
                params=(xc1,yc1,m,n,a1,b1,eps1,freq,mode)
                Qw1[i,i] = np.sum(eh(xp_1,yp_1,params)*dS1)

            for i in range(Nw2):
                m,n,mode = modes2[i]
                params=(xc2,yc2,m,n,a2,b2,eps2,freq,mode)
                Qw2[i,i] = np.sum(eh(xp_2,yp_2,params)*dS2)

            if xt2<xt1 or yt2<yt1:
                Xw1=0
                Xw2=0
                print("hh")
            else:
                Xw1 = np.matrix(np.zeros((Ns,Nw1),dtype=complex))
                for i in range(len(modes3)):
                    m3,n3,mode3 = modes3[i]
                    params2=(xt1,yt1,m3,n3,xt2-xt1,yt2-yt1,0.5*(eps1+eps2),freq,mode3)
                    for j in range(len(modes1)):
                        m1,n1,mode1 = modes1[j]
                        params1=(xc1,yc1,m1,n1,a1,b1,eps1,freq,mode1)
                        Xw1[i,j] = np.sum(exh(xp_3,yp_3,params1,params2)*dS3)
                Xw2 = np.matrix(np.zeros((Ns,Nw2),dtype=complex))
                for i in range(len(modes3)):
                    m3,n3,mode3 = modes3[i]
                    params2=(xt1,yt1,m3,n3,xt2-xt1,yt2-yt1,0.5*(eps1+eps2),freq,mode3)
                    for j in range(len(modes2)):
                        m2,n2,mode2 = modes2[j]
                        params1=(xc2,yc2,m2,n2,a2,b2,eps2,freq,mode2)
                        Xw2[i,j] = np.sum(exh(xp_3,yp_3,params1,params2)*dS3)

            F = 2*(Xw1*Qw1.I*Xw1.T+Xw2*Qw2.I*Xw2.T).I
            S11 = Qw1.I*Xw1.T*F*Xw1-Iw1
            S12 = Qw1.I*Xw1.T*F*Xw2
            S21 = Qw2.I*Xw2.T*F*Xw1
            S22 = Qw2.I*Xw2.T*F*Xw2-Iw2
            SP[:Nw1,:Nw1] = S11
            SP[:Nw1,Nw1:] = S12
            SP[Nw1:,:Nw1] = S21
            SP[Nw1:,Nw1:] = S22
            SPARAMS.append(SP)
        return SPARAMS, modes1, modes2

if __name__=="__main__":
    xc1=0
    yc1=0
    xc2=0
    yc2=0
    a1=3.1e-3
    b1=1.55e-3
    a2=3.1e-3
    # b2=0.55e-3
    b2=1.55e-3
    eps1=1
    eps2=1
    aperture1=(xc1,yc1,a1,b1,eps1)
    aperture2=(xc2,yc2,a2,b2,eps2)
    aperture3=(1.05e-3,2.05e-3,0,1.55e-3)
    # freqs=np.linspace([65e9,90e9,26])
    freqs=[77e9]
    solver=MM_RWG(aperture1,aperture2,aperture3)
    Sparams, modes1, modes2 = solver.simulate(freqs)
    cc1=modes1.index((1,0,1))
    print(cc1)
    print(Sparams[0][cc1,cc1])
    # import matplotlib.pyplot as plt
    # plt.plot(freqs, 20*np.log10(np.abs(np.array(RL))))
    # plt.grid()
    # plt.show()



#-*-coding:utf-8-*-
"""
Network Parameters
"""
import numpy as np
import operator as op
from numpy.lib.scimath import sqrt as csqrt
from functools import reduce

def ideal_nport(N):
    """
    S-parameters of ideal N-port junction with equal reference impedances at all ports
    """
    S = np.matrix(np.ones((N,N),dtype=float))
    a=(2-N)/N
    b=2/N
    S=b*S
    for i in range(N):
        S[i,i]=a
    return S

def ideal_gyrator():
    """
    S-parameters of ideal gyrator
    """
    return np.matrix([[0,-1],[1,0]])

def ideal_coupled_line(Ze, Zo, Te, To, Z0):
    """
    S-parameters of ideal coupled line
    Te and To in radian
    3----------4
    1----------2
    """
    De = 2*Ze*Z0*np.cos(Te)+(Ze**2+Z0**2)*np.sin(Te)*1j
    Do = 2*Zo*Z0*np.cos(To)+(Zo**2+Z0**2)*np.sin(To)*1j
    Ye = Ze*Z0/De
    Yo = Zo*Z0/Do
    Xe = (Ze**2-Z0**2)*np.sin(Te)*1j/2/De
    Xo = (Zo**2-Z0**2)*np.sin(To)*1j/2/Do
    S  = np.matrix(np.ones((4,4),dtype=complex))
    S[0,0]=S[1,1]=S[2,2]=S[3,3]=Xe+Xo
    S[0,1]=S[1,0]=S[2,3]=S[3,2]=Ye+Yo
    S[0,2]=S[2,0]=S[1,3]=S[3,1]=Xe-Xo
    S[0,3]=S[3,0]=S[1,2]=S[2,1]=Ye-Yo
    return S

def ideal_amp(G):
    """
    S-parameters of an ideal amplifier/isolator
    G is voltage gain, no reflection, infinite isolation
    """
    return np.matrix([[0, 0], [G, 0]])

def ideal_att(G, db=False):
    """
    S-parameters of an ideal attenuator
    G is voltage gain (<1), no reflection
    """
    if db:
        G = 10 ** ( -np.abs(G) / 20 )
    return np.matrix([[0.00001, G], [G, 0.00001]])

def circulator():
    """
    S-parameters of an ideal circulator (circulation direction 1⇒2⇒3)
    """
    return np.matrix([[0, 0, 1.], [1., 0, 0], [0, 1., 0]])

def shunt_z(Z):
    """
    ABCD parameters of shunt impedance
    """
    return np.matrix([[1., 0], [(1./Z), 1.]])

def series_z(Z):
    """
    ABCD parameters of series impedance
    """
    return np.matrix([[1., Z], [0, 1.]])

def jinv(J):
    """
    ABCD parameters of J - inverter
    """
    return np.matrix([[0, 1.0j/J], [1.0j * J, 0]])

def jinv_lumped(X):
    """
    ABCD parameters of J - inverter produced by 3 inductors in Tee form.
    """
    return cascade_networks([shZ( - X), seZ(X), shZ( - X)])

def kinv(K):
    """
    ABCD parameters of k - inverter
    """
    return np.matrix([[0, 1.0j * K], [1.0j/K, 0]])

def tline(Zo, theta):
    """
    ABCD parameters of ideal transmission line,  theta = radian
    """
    return np.matrix([[np.cos(theta), 1.0j * Zo * np.sin(theta)], [1.0j/Zo * np.sin(theta), np.cos(theta)]])

def tline_lossy(Zo, gamma, length):
    """
    ABCD parameters of lossy transmission line,  gamma = complex propagation constant
    """
    return np.matrix([[np.cosh(gamma * length), Zo * np.sinh(gamma * length)],
                      [1.0/Zo * np.sinh(gamma * length), np.cosh(gamma * length)]])

def tline_list(Zo, theta):
    """
    ABCD parameters of ideal transmission line,  theta = radian
    """
    return tuple([np.cos(theta), 1.0j * Zo * np.sin(theta), 1.0j/Zo * np.sin(theta), np.cos(theta)])

def transformer(N):
    """
    ABCD parameters of ideal transformer (1:N)
    """
    return np.matrix([[1./N, 0], [0, N]]
                     )
def t_network(Zs1, Zp, Zs2):
    """
    ABCD parameters of Tee network
    """
    return np.matrix([[1 + Zs1/Zp, Zs1 + Zs2 + Zs1 * Zs2/Zp], [1./Zp, 1. + Zs2/Zp]])

def pi_network(Zp1, Zs, Zp2):
    """
    ABCD parameters of Pi network
    """
    return np.matrix([[1 + Zs/Zp2, Zs], [1./Zp1 + 1./Zp2 + Zs/Zp1/Zp2, 1 + Zs/Zp1]])

def abcd2y(M):
    """
    ABCD parameters to Y-Parameters conversion
    """
    a, b, c, d = M[0, 0], M[0, 1], M[1, 0], M[1, 1]
    return np.matrix([[d/b, (b * c - a * d)/b], [ -1./b, a/b]])

def y2abcd(M):
    """
    Y-Parameters to ABCD parameters conversion
    """
    y11, y12, y21, y22 = M[0, 0], M[0, 1], M[1, 0], M[1, 1]
    return np.matrix([[ -y22/y21,  -1./y21], [(y12 * y21 - y11 * y22)/y21,  -y11/y21]])

def t2s(M):
    """
    Transfer scattering parameters to S-Parameters conversion
    According to definition [b1,a1]=T.[a2,b2]
    Ref: https://en.wikipedia.org/wiki/Scattering_parameters#Scattering_transfer_parameters
    """
    t11, t12, t21, t22 = M[0, 0], M[0, 1], M[1, 0], M[1, 1]
    delta = t11*t22-t12*t21
    return np.matrix([[ t12/t22,  delta/t22], [1./t22,  -t21/t22]])

def t2s_list(M):
    """
    Transfer scattering parameters to S-Parameters conversion
    According to definition [b1,a1]=T.[a2,b2]
    Ref: https://en.wikipedia.org/wiki/Scattering_parameters#Scattering_transfer_parameters
    """
    t11, t12, t21, t22 = M[0], M[1], M[2], M[3]
    delta = t11*t22-t12*t21
    return [ t12/t22,  delta/t22, 1./t22,  -t21/t22]

def s2t(M):
    """
    S-Parameters to Transfer scattering parameters conversion
    According to definition [b1,a1]=T.[a2,b2]
    Ref: https://en.wikipedia.org/wiki/Scattering_parameters#Scattering_transfer_parameters
    """
    s11, s12, s21, s22 = M[0, 0], M[0, 1], M[1, 0], M[1, 1]
    delta = s11*s22-s12*s21
    return np.matrix([[ -delta/s21,  s11/s21], [-s22/s21,  1./s21]])

def s2t_list(M):
    """
    S-Parameters to Transfer scattering parameters conversion
    According to definition [b1,a1]=T.[a2,b2]
    Ref: https://en.wikipedia.org/wiki/Scattering_parameters#Scattering_transfer_parameters
    """
    s11, s12, s21, s22 = M[0], M[1], M[2], M[3]
    delta = s11*s22-s12*s21
    return [ -delta/s21,  s11/s21, -s22/s21,  1./s21]

def abcd2z(M):
    """
    ABCD parameters to Z-Parameters conversion
    """
    a, b, c, d = M[0, 0], M[0, 1], M[1, 0], M[1, 1]
    return np.matrix([[a/c, ( - b * c + a * d)/c], [1./c, d/c]])

def z2abcd(M):
    """
    Z-Parameters to ABCD parameters conversion
    """
    z11, z12, z21, z22 = M[0, 0], M[0, 1], M[1, 0], M[1, 1]
    return np.matrix([[z11/z21, (z11 * z22 - z21 * z12)/z21],
                      [1./z21, z22/z21]])

def abcd2h(M):
    """
    ABCD parameters to H-Parameters conversion
    """
    a, b, c, d = M[0, 0], M[0, 1], M[1, 0], M[1, 1]
    return np.matrix([[b/d, ( - b * c + a * d)/d],
                      [-1./d, c/d]])

def h2abcd(M):
    """
    H-Parameters to ABCD parameters conversion
    """
    h11, h12, h21, h22 = M[0, 0], M[0, 1], M[1, 0], M[1, 1]
    return np.matrix([[h12 - h22 * h11 / h21, -h11 / h21],
                      [-h22 / h21, -1. / h21]])

def h2s(M, Zo=50.0):
    """
    H-Parameters to S parameters conversion. Valid for real Zo value.
    """
    h11, h12, h21, h22 = M[0, 0], M[0, 1], M[1, 0], M[1, 1]
    delta = (h11 + Zo) * (h22 * Zo + 1.0) - h12 * h21 * Zo
    return np.matrix([[((h11 - Zo) * (h22 * Zo + 1.0) - h12 * h21 * Zo) / delta, 2 * h12 * Zo / delta],
                      [-2 * h12 * Zo / delta, ((h11 + Zo) * (1.0 - h22 * Zo) + h12 * h21 * Zo) / delta]])

def h2y(M):
    """
    H-Parameters to Y-parameters conversion for 2-port networks.
    """
    h11, h12, h21, h22 = M[0, 0], M[0, 1], M[1, 0], M[1, 1]
    return np.matrix([[1/h11, -h12 / h11],
                      [h21 / h11, (h22 - h21 * h12 / h11)]])

def s2h(M, Zo=50.0):
    """
    S-Parameters to H parameters conversion. Valid for real Zo value.
    """
    s11, s12, s21, s22 = M[0, 0], M[0, 1], M[1, 0], M[1, 1]
    delta = (1. - s11) * (1. + s22) + s12 * s21
    return np.matrix([[Zo * ((1. + s11) * (1. + s22) - s12 * s21) / delta, 2 * s12 / delta],
                      [-2 * s21 / delta, 1.0 / Zo * ((1. - s11) * (1. - s22) - s12 * s21) / delta]])

def s2g(M, Zo=50.0):
    """
    S-Parameters to G parameters conversion. Valid for real Zo value.
    """
    H = s2h(M, Zo)
    return H.I

def abcd2g(M):
    """
    S-Parameters to G parameters conversion.
    """
    H = abcd2h(M)
    return H.I

def abcd2s(M, Zo=50.0):
    """
    ABCD parameters to S-Parameters conversion. Valid for real Zo value.
    """
    a, b, c, d = M[0, 0], M[0, 1], M[1, 0], M[1, 1]
    s11 = ((a + b/Zo - c * Zo - d)/(a + b/Zo + c * Zo + d))
    s12 = 2. * (a * d - b * c)/(a + b/Zo + c * Zo + d)
    s21 = (2./(a + b/Zo + c * Zo + d))
    s22 = (( - a + b/Zo - c * Zo + d)/(a + b/Zo + c * Zo + d))
    return np.matrix( [[s11, s12], [s21, s22]] )

def abcd2s_list(M, Zo=50.0):
    """
    ABCD parameters to S-Parameters conversion. Valid for real Zo value.
    """
    a, b, c, d = M[0], M[1], M[2], M[3]
    s11 = ((a + b/Zo - c * Zo - d)/(a + b/Zo + c * Zo + d))
    s12 = 2. * (a * d - b * c)/(a + b/Zo + c * Zo + d)
    s21 = (2./(a + b/Zo + c * Zo + d))
    s22 = (( - a + b/Zo - c * Zo + d)/(a + b/Zo + c * Zo + d))
    return [s11, s12, s21, s22]

def s2abcd(M, Z=(50.0, 50.0)):
    """
    S-Parameters to ABCD parameters conversion. Valid for real Z values.

    Args:
        Z(2-tuple, optional): reference impedance tuple ( Z1, Z2 )
    """
    if isinstance(Z, (int, float)):
        Zo1 = Zo2 = Z
    else:
        Zo1, Zo2 = tuple(Z)
    s11, s12, s21, s22 = M[0, 0], M[0, 1], M[1, 0], M[1, 1]
    a = ((Zo1 + s11 * Zo1) * (1. - s22) + s12 * s21 * Zo1)/(2. * s21 * csqrt(Zo1 * Zo2))
    b = ((Zo1 + s11 * Zo1) * (Zo2 + s22 * Zo2) - s12 * s21 * Zo1 * Zo2)/(2. * s21 * csqrt(Zo1 * Zo2))
    c = ((1. - s11) * (1. - s22) - s12 * s21)/(2. * s21 * csqrt(Zo1 * Zo2))
    d = ((1. - s11) * (Zo2 + s22 * Zo2) + s12 * s21 * Zo2)/(2. * s21 * csqrt(Zo1 * Zo2))
    return np.matrix([[a, b], [c, d]])

def abcd2t(M, Zo=50.0):
    """
    ABCD parameters to T-Parameters conversion. Valid for real impedances.

    ABCD: [V1 I1]=ABCD*[V2 -I2]
    Pseudo-Wave or Power-Wave? Don't use for complex impedances.
    """
    X = abcd2s(M, Zo)
    return s2t(X)

def abcd_change_ports(M):
    """
    Switching ports of ABCD parameters
    """
    M = M.I
    np.matrix([[M[0, 0],  - M[0, 1]], [ - M[1, 0], M[1, 1]]])

def t2abcd(M, Z=(50.0,50.0)):
    """
    T-parameters to ABCD parameters conversion
    """
    X = t2s(M)
    return s2abcd(X, Z)

def snp2smp(SM,ports):
    """
    This method changes the port numbering of the network
    port j of new network corresponds to ports[j] in old network
    if the length of "ports" argument is lower than number of ports, remaining ports are terminated with current reference impedances and number of ports are reduced.
    """
    ports = [x-1 for x in ports]
    ixgrid = np.ix_(ports, ports)
    return SM[ixgrid]

def cascade_networks(networks):
    """
    Cascading 2-port Networks,  input and output is ABCD matrices of networks
    """
    return reduce(op.mul, networks)

def parallel_networks(networks):
    """
    Paralleling 2-port Networks,  input and output  is ABCD matrices of networks
    """
    ymatrices = [abcd2y(M) for M in networks]
    return y2abcd(reduce(op.add, ymatrices))

def series_networks(networks):
    """
    Series Connection of Networks (reference pins of 1. network is connected to alive pins of 2. network),  input and output  is ABCD matrices of networks
    """
    zmatrices = [abcd2z(M) for M in networks]
    return z2abcd(reduce(op.add, zmatrices))

def s_normalize_pseudo(S, Zold, Znew):
    """
    Zold,  Znew port_sayisi uzunlugunda dizilerdir
    Pseudo-Wave icin
    """
    ps = len(Zold)
    A = np.matrix(np.zeros((ps, ps)), dtype = complex)
    gamma = np.matrix(np.zeros((ps, ps)), dtype = complex)
    imp_yeni = []
    for i in range(ps):
        imp_yeni.append(Znew[i])
        z = Zold[i]
        ri = ((imp_yeni[i] - z)/(imp_yeni[i] + z))
        gamma[i, i] = ri
        A[i, i] = csqrt(z.real/imp_yeni[i].real) * abs(imp_yeni[i]/z) * 2 * z/(imp_yeni[i] + z)
    return A.I * (S - gamma) * (np.matrix(np.eye(ps)) - gamma * S).I * A

def s_normalize_power(S, Zold, Znew):
    """
    Zold,  Znew port_sayisi uzunlugunda dizilerdir
    Power-Wave icin
    Reference: Article, “Multiport conversions between S, Z, Y, h, ABCD, and T parameters”
    """
    ps = len(Zold)
    A = np.matrix(np.zeros((ps, ps)), dtype = complex)
    gamma = np.matrix(np.zeros((ps, ps)), dtype = complex)
    imp_yeni = []
    for i in range(ps):
        imp_yeni.append(Znew[i])
        z = Zold[i]
        gamma[i, i] = ((imp_yeni[i] - z)/(imp_yeni[i] + z.conj()))
        A[i, i] = 2*csqrt(z.real)*csqrt(imp_yeni[i].real) /(imp_yeni[i].conj() + z)
    return A.I * (S - gamma.conj()) * (np.matrix(np.eye(ps)) - gamma * S).I * A.conj()

def s_phase_deembed(S, phase):
    """
    S-parameter deembedding
    S is numpy.matrix NxN
    phase, deembedding phase for each port in radian. Positive phase is deembedding into the circuit
    """
    phase_matrix=np.exp(1j*np.matrix(np.diag(phase)))
    return phase_matrix*S*phase_matrix

def connect_2_ports(Smatrix,k,m):
    """ Port-m is connected to port-k and both ports are removed
    Reference: QUCS technical.pdf, S-parameters in CAE programs, p.29
    """
    k,m=min(k,m),max(k,m)
    k=k-1
    m=m-1
    ps=np.shape(Smatrix)[0]
    Sm=np.matrix(np.zeros((ps-2,ps-2),dtype=complex)) # output matrix
    S = Smatrix
    for i in range(ps-2):
        ii=i+(i>=k)+(i>=(m-1))
        for j in range(ps-2):
            jj=j+(j>=k)+(j>=(m-1))
            temp = S[k,jj]*S[ii,m]*(1-S[m,k])+S[m,jj]*S[ii,k]*(1-S[k,m])+S[k,jj]*S[m,m]*S[ii,k]+S[m,jj]*S[k,k]*S[ii,m]
            Sm[i,j] = S[ii,jj] + temp/((1-S[m,k])*(1-S[k,m])-S[k,k]*S[m,m])
    return Sm

def connect_network_1_conn_retain(Smatrix,EX,k,m):
    ideal3port = ideal_nport(3)
    EX = connect_network_1_conn(EX,ideal3port,m,1)
    psex = EX.shape[0]
    sonuc = connect_network_1_conn(Smatrix,EX,k,psex)
    return sonuc

def connect_network_1_conn(Smatrix,EX,k,m):
    """ Port-m of EX circuit is connected to port-k of this circuit
    Remaining ports of EX are added to the port list of this circuit in order.
    Reference: QUCS technical.pdf, S-parameters in CAE programs, p.29
    """
    S = Smatrix
    k=k-1
    m=m-1
    ps1=np.shape(S)[0]
    ps2=np.shape(EX)[0]
    ps=ps1+ps2-2
    Sm=np.matrix(np.ones((ps,ps),dtype=complex))
    for i in range(ps1-1):
        ii=i+(i>(k-1))
        for j in range(ps1-1):
            jj=j+(j>(k-1))
            # index = (i-1)*ps+(j-1)
            Sm[i,j] = S[ii,jj]+S[k,jj]*EX[m,m]*S[ii,k]/(1.0-S[k,k]*EX[m,m])
    for i in range(ps2-1):
        ii=i+(i>(m-1))
        for j in range(ps1-1):
            jj=j+(j>(k-1))
            # index = (i+ps1-1-1)*ps+(j-1)
            Sm[i+ps1-1,j] = S[k,jj] * EX[ii,m] / (1.0 - S[k,k] * EX[m,m])
    for i in range(ps1-1):
        ii=i+(i>(k-1))
        for j in range(ps2-1):
            jj=j+(j>(m-1))
            # index = (i-1)*ps+(j+ps1-1-1)
            Sm[i,j+ps1-1] = EX[m,jj] * S[ii,k] / (1.0 - EX[m,m] * S[k,k])
    for i in range(ps2-1):
        ii=i+(i>(m-1))
        for j in range(ps2-1):
            jj=j+ (j>(m-1))
            # index = (i+ps1-1-1)*ps+(j+ps1-1-1)
            Sm[i+ps1-1,j+ps1-1] = EX[ii,jj]+EX[m,jj]*S[k,k]*EX[ii,m]/(1.0-EX[m,m]*S[k,k])
    return Sm

def connect_2_ports_list(Smatrix,conns):
    """ Short circuit ports together one-to-one. Short circuited ports are removed.
    Ports that will be connected are given as tuples in list conn
    i.e. conn=[(p1,p2),(p3,p4),..]
    The order of remaining ports is kept.
    Reference: QUCS technical.pdf, S-parameters in CAE programs, p.29
    """
    conns = [[i,j] for i,j in conns]
    for i in range(len(conns)):
        k,m = conns[i]
        Smatrix = connect_2_ports(Smatrix, k,m)
        for j in range(i+1,len(conns)):
            conns[j][0]=conns[j][0]-(conns[j][0]>k)-(conns[j][0]>m)
            conns[j][1]=conns[j][1]-(conns[j][1]>k)-(conns[j][1]>m)
    return Smatrix

def connect_2_ports_retain(Smatrix,k,m):
    """ Port-m and Port-k are joined to a single port.
    New port becomes the last port of the circuit.
    Reference: QUCS technical.pdf, S-parameters in CAE programs, p.29
    """
    ideal3port = 1/3*np.matrix([[-1,2,2],[2,-1,2],[2,2,-1]])
    ps = np.shape(Smatrix)[0]
    k,m=min(k,m),max(k,m)
    Sm = connect_network_1_conn(Smatrix,ideal3port,m,1)
    Sm = connect_2_ports(Sm,k,ps)
    return Sm

if __name__=="__main__":
    print(ideal_nport(4))



import numpy as np

def refinearray(inp):
    i=0
    if len(inp)<2:
        return inp
    while 1:
        if np.abs(inp[i]-inp[i+1])/np.sum(inp)<1e-5:
            inp.pop(i+1)
        else:
            i=i+1
        if (i>=len(inp)-1):
            break
    return inp

class subregion:
    def __init__(self):
        self.start=0
        self.end=0
        self.r=0
        self.mincell=0
        self.maxcell=0
        self.startindex=0
        self.endindex=0
        self.Ncell=0
        self.shape=0  #shape: uniform=1, FC=2, CF=3, FCF=4
        #startindex bolgenin icindeki ilk hucreyi isaret eder, endindex bolgenin disindaki ilk hucreyi isaret eder.

class CartesianGrid():
    def __init__(self):
        self.sizex=0
        self.sizey=0
        self.sizez=0
        self.sensitivity=1e-5
        self.number_of_abc_layers_x1=0
        self.number_of_abc_layers_x2=0
        self.number_of_abc_layers_y1=0
        self.number_of_abc_layers_y2=0
        self.number_of_abc_layers_z1=0
        self.number_of_abc_layers_z2=0
        self.dx=[]
        self.dy=[]
        self.dz=[]
        self.KritikNoktalar_x=[]
        self.KritikNoktalar_y=[]
        self.KritikNoktalar_z=[]
        self.delta=0
        self.deltamin=0
        self.unit ="m"
        self.maxratio=1.2

    def findsubregions(self, kritikx,kritiky,kritikz):

        kritikx.sort()
        kritiky.sort()
        kritikz.sort()

        refinearray(kritikx)
        refinearray(kritiky)
        refinearray(kritikz)

        self.KritikNoktalar_x=kritikx
        self.KritikNoktalar_y=kritiky
        self.KritikNoktalar_z=kritikz

        self.sizex=len(kritikx)
        self.sizey=len(kritiky)
        self.sizez=len(kritikz)

        subx = []
        suby = []
        subz = []

        for i in range(self.sizex-1):
            sr = subregion()
            sr.start = kritikx[i]
            sr.end = kritikx[i+1]
            sr.shape = 1
            subx.append(sr)

        for i in range(self.sizey-1):
            sr = subregion()
            sr.start = kritiky[i]
            sr.end = kritiky[i+1]
            sr.shape = 1
            suby.append(sr)

        for i in range(self.sizez-1):
            sr = subregion()
            sr.start = kritikz[i]
            sr.end = kritikz[i+1]
            sr.shape = 1
            subz.append(sr)

        self.subx=subx
        self.suby=suby
        self.subz=subz

    def optimizegrids(self):
        maxratio=self.maxratio
        sublists = [self.subx,self.suby,self.subz]
        gridler = [self.dx,self.dy,self.dz]
        policy=0
        for j in range(3):
            sublist=sublists[j]
            dd=gridler[j]
            if len(sublist)<2:
                return
            if dd[sublist[0].endindex-1]>maxratio*dd[sublist[0].endindex]:
                self.customgrid(maxratio*dd[sublist[0].endindex], dd[sublist[0].startindex], 0, maxratio, 1, j, policy)
            for i in range(1,len(sublist)-1):
                if dd[sublist[i].startindex]>maxratio*dd[sublist[i].startindex-1] and dd[sublist[i].endindex-1]>maxratio*dd[sublist[i].endindex]:
                    self.customgrid(maxratio*np.min((dd[sublist[i].startindex-1],dd[sublist[i].endindex])), dd[sublist[i].startindex], i, maxratio, 2, j, policy)
                elif dd[sublist[i].startindex]>maxratio*dd[sublist[i].startindex-1]:
                    self.customgrid(maxratio*dd[sublist[i].startindex-1], dd[sublist[i].endindex-1], i, maxratio, 0, j, policy)
                elif dd[sublist[i].endindex-1]>maxratio*dd[sublist[i].endindex]:
                    self.customgrid(maxratio*dd[sublist[i].endindex], maxratio*dd[sublist[i].startindex-1], i, maxratio, 1, j, policy)
            if dd[sublist[-1].startindex]>maxratio*dd[sublist[-1].startindex-1]:
                self.customgrid(maxratio*dd[sublist[-1].startindex-1], dd[sublist[-1].endindex-1], -1, maxratio, 0, j, policy)


    def nu_grid(self, deltax, deltay, deltaz):
        dx=[]
        dy=[]
        dz=[]
        sizex = self.sizex
        sizey = self.sizey
        sizez = self.sizez
        subx = self.subx
        suby = self.suby
        subz = self.subz
        temp=0
        for i in range(self.sizex-1):
            grids=self.ugrid(deltax,subx[i].end-subx[i].start)
            subx[i].Ncell=len(grids)
            subx[i].startindex=temp
            temp+=subx[i].Ncell
            subx[i].endindex=temp
            dx.extend(grids)
            subx[i].mincell=grids[0]
            subx[i].maxcell=grids[0]
        temp=0
        for i in range(self.sizey-1):
            grids=self.ugrid(deltay,suby[i].end-suby[i].start)
            suby[i].Ncell=len(grids)
            suby[i].startindex=temp
            temp+=suby[i].Ncell
            suby[i].endindex=temp
            dy.extend(grids)
            suby[i].mincell=grids[0]
            suby[i].maxcell=grids[0]
        temp=0
        for i in range(self.sizez-1):
            grids=self.ugrid(deltaz,subz[i].end-subz[i].start)
            subz[i].Ncell=len(grids)
            subz[i].startindex=temp
            temp+=subz[i].Ncell
            subz[i].endindex=temp
            dz.extend(grids)
            subz[i].mincell=grids[0]
            subz[i].maxcell=grids[0]
        self.dx = dx
        self.dy = dy
        self.dz = dz
        self.subx = subx
        self.suby = suby
        self.subz = subz


    def set_number_of_abc_layers(self, x1, x2, y1, y2, z1, z2):
        if (x1>=0): self.number_of_abc_layers_x1=x1
        if (x2>=0): self.number_of_abc_layers_x2=x2
        if (y1>=0): self.number_of_abc_layers_y1=y1
        if (y2>=0): self.number_of_abc_layers_y2=y2
        if (z1>=0): self.number_of_abc_layers_z1=z1
        if (z2>=0): self.number_of_abc_layers_z2=z2

    def writetofile(self, dosya):
        pass
    def getsubregion(self, i):
        x1=self.subx[i].start
        x2=self.subx[i].end
        y1=self.suby[i].start
        y2=self.suby[i].end
        z1=self.subz[i].start
        z2=self.subz[i].end
        return x1, x2, y1, y2, z1, z2

    def getsubregion_p(self, j, i):
        if (j==0):
            return (self.subx[i].start,self.subx[i].end)
        elif (j==1):
            return (self.suby[i].start,self.suby[i].end)
        elif (j==2):
            return (self.subz[i].start,self.subz[i].end)
        return (0,0)

    def customgrid(self, dmin, dmax, index, rmax, shape, axis, policy):
        """
        Args:
            index (int): subregion numarasi
            rmax (float):      maximum ratio between neighboring grids
            shape (int):0- fine-coarse
                        1- coarse-fine
                        2- fine-coarse-fine
                        3- uniform
            policy (int):   0- minimum number of grids
                            1- minimum ratio, smoother transition
        """

        dx = self.dx
        dy = self.dy
        dz = self.dz
        sizex = self.sizex
        sizey = self.sizey
        sizez = self.sizez
        subx = self.subx
        suby = self.suby
        subz = self.subz
        dmin, dmax = np.min((dmin,dmax)), np.max((dmin,dmax))
        if (axis==0):
            if (index>=sizex-1):
                print("hata")
                return

            bas=subx[index].startindex
            son=subx[index].endindex
            length=subx[index].end-subx[index].start
            grids = self.nugrid(dmin,dmax,rmax,length,shape,policy)
            subx[index].Ncell=len(grids)
            subx[index].endindex=len(grids)+bas
            for i in range(index+1, sizex-1):
                subx[i].startindex+=len(grids)-(son-bas)
                subx[i].endindex+=len(grids)-(son-bas)

            del dx[bas:son]
            dx[bas:bas] = grids
            subx[index].mincell=grids[0]
            subx[index].maxcell=grids[len(grids)-1]

        elif (axis==1):
            if (index>=sizey-1):
                print("hata")
                return

            bas=suby[index].startindex
            son=suby[index].endindex
            length=suby[index].end-suby[index].start
            grids=self.nugrid(dmin,dmax,rmax,length,shape,policy)
            suby[index].Ncell=len(grids)
            suby[index].endindex=len(grids)+bas
            for i in range(index+1,sizey-1):
                suby[i].startindex+=len(grids)-(son-bas)
                suby[i].endindex+=len(grids)-(son-bas)

            del dy[bas:son]
            dy[bas:bas] = grids
            suby[index].mincell=grids[0]
            suby[index].maxcell=grids[len(grids)-1]


        elif (axis==2):
            if (index>=sizez-1):
                print("hata")
                return

            bas=subz[index].startindex
            son=subz[index].endindex
            length=subz[index].end-subz[index].start
            grids=self.nugrid(dmin,dmax,rmax,length,shape,policy)
            subz[index].Ncell=len(grids)
            subz[index].endindex=len(grids)+bas
            for i in range(index+1,sizez-1):
                subz[i].startindex+=len(grids)-(son-bas)
                subz[i].endindex+=len(grids)-(son-bas)

            del dz[bas:son]
            dz[bas:bas] = grids
            subz[index].mincell=grids[0]
            subz[index].maxcell=grids[len(grids)-1]

        self.dx = dx
        self.dy = dy
        self.dz = dz
        self.subx = subx
        self.suby = suby
        self.subz = subz

    def reshapegrid(self, delta, index, rmax, shape, axis, policy):
        #index: subregion numarasi
        dx = self.dx
        dy = self.dy
        dz = self.dz
        sizex = self.sizex
        sizey = self.sizey
        sizez = self.sizez
        subx = self.subx
        suby = self.suby
        subz = self.subz
        tip=0
        if (axis==0):
            if (index==0):
                d1=delta
                d2=dx[subx[0].endindex]
            elif (index==(sizex-2)):
                d2=delta
                d1=dx[subx[sizex-2].startindex-1]
            elif ((index>0) and (index<(sizex-2))):
                d1=dx[subx[index].startindex-1]
                d2=dx[subx[index].endindex]
            else:
                print("Error in index number ")

        elif (axis==1):
            if (index==0):
                d1=delta
                d2=dy[suby[0].endindex]

            elif (index==(sizey-2)):
                d2=delta
                d1=dy[suby[sizey-2].startindex-1]

            elif ((index>0) and (index<(sizey-2))):
                d1=dy[suby[index].startindex-1]
                d2=dy[suby[index].endindex]

            else:
                print("Error in index number ")

        elif (axis==2):
            if (index==0):
                d1=delta
                d2=dz[subz[0].endindex]

            elif (index==(sizez-2)):
                d2=delta
                d1=dz[subz[sizez-2].startindex-1]

            elif ((index>0) and (index<(sizez-2))):
                d1=dz[subz[index].startindex-1]
                d2=dz[subz[index].endindex]

            else:
                print("Error in index number ")

        if (d2<d1):
            tip=1
        else:
            tip=0
        self.customgrid(d1,d2,index,rmax,tip,axis,policy)

    def sumz(self):
        return np.sum(dz)

    def get_grids_sub_i(self, i, axis):
        if (axis==0):
            return self.dx[self.subx[i].startindex:self.subx[i].endindex]
        elif (axis==1):
            return self.dy[self.suby[i].startindex:self.suby[i].endindex]
        elif (axis==2):
            return self.dz[self.subz[i].startindex:self.subz[i].endindex]

    def get_subregion_index_of_point(self, coor, axis):
        i=0
        if (axis==0):
            while (i<self.sizex):
                if ((coor>=self.subx[i].start) and (coor<=self.subx[i].end)):
                    return i
                else:
                    i+=1

        elif (axis==1):
            while (i<self.sizey):
                if ((coor>=self.suby[i].start) and (coor<=self.suby[i].end)):
                    return i
                else:
                    i+=1

        elif (axis==2):
            while (i<self.sizez):
                if ((coor>=self.subz[i].start) and (coor<=self.subz[i].end)):
                    return i
                else:
                    i+=1

    def loadfromfile(self, dosya):
        pass

    def subregionparams(self, c, grid):
        if (grid==0):
            a = self.subx[c].mincell
            b = self.subx[c].maxcell
            shape = self.subx[c].shape
        elif (grid==1):
            a = self.suby[c].mincell
            b = self.suby[c].maxcell
            shape = self.suby[c].shape
        elif (grid==2):
            a = self.subz[c].mincell
            b = self.subz[c].maxcell
            shape = self.subz[c].shape
        return a, b, shape

    def return_subregion_params(self, grid, index):
        if (grid==0):
            a = self.subx[index].mincell
            b = self.subx[index].maxcell
        elif (grid==1):
            a = self.suby[index].mincell
            b = self.suby[index].maxcell
        elif (grid==2):
            a = self.subz[index].mincell
            b = self.subz[index].maxcell
        return (a,b)

    def ugrid(self, delta, length, parite=0):
        N = int(np.ceil(length/delta))
        if ((parite==-1) and (N%2==0)):
            N = N+1
        if ((parite==1) and (N%2==1)):
            N = N+1
        if (N==0):
            N=1
        grids = [length/N]*N
        return grids

    def nugrid(self, delta1, delta2, rmax, length, shape,  policy=0,cc=0.99):
        """
        Nonuniform grid generation.

        Args:
            rmax (float): Maximum ratio between neighboring grids
            length (float):     total length
            shape (int):    0- fine-coarse
                            1- coarse-fine
                            2- fine-coarse-fine
                            3- uniform
            policy (int):   0- minimum number of grids
                            1- minimum ratio, smoother transition
        """
        r=rmax

        if (shape==3):
            grids=self.ugrid(delta1,length)
        elif  (shape==2):
            grids=self.nugrid(delta1,delta2,rmax,length/2.0,0, policy)
            temp=self.nugrid(delta1,delta2,rmax,length/2.0,1, policy)
            grids = grids + temp
        elif  (shape==0 or shape==1):
            grids=[]
            deltamin=np.min([delta1,delta2])
            deltamax=np.max([delta1,delta2])
            if (policy==1):
                r=(length-deltamin)/(length-deltamax)
                N=round(float(np.ceil(np.log(deltamax/deltamin)/np.log(r))))
                while(1):
                    temp3=deltamin*(r**(N+1)-1)/(r-1)
                    if (np.abs(temp3/length-1)<self.sensitivity):
                        break
                    elif (temp3>length):
                        r=r*cc
                    elif (temp3<length):
                        r=r*(2.0-cc)
                error = length-temp3
                grids = grids + [deltamin*np.power(r,i) for i in range(N+1)]
                grids[-1] += error
            elif (policy==0):
                N=1+int(np.ceil(np.log(deltamax/deltamin)/np.log(rmax)))
                temp2=deltamin*(np.power(rmax,N)-1)/(rmax-1)
                if (temp2<length):
                    N3=int(np.ceil((length-temp2)/deltamax))
                    while(1):
                        temp3=deltamin*(np.power(r,N)-1)/(r-1)+N3*deltamin*np.power(r,N-1)
                        if (np.abs(temp3/length-1)<self.sensitivity):
                            break
                        elif (temp3>length):
                            r=r*cc
                        elif (temp3<length):
                            r=r*(2.0-cc)
                    error = length-temp3
                    grids = grids + [deltamin*np.power(r,i) for i in range(N)]
                    grids[-1] += error
                    grids = grids + [deltamin*np.power(r,N-1) for i in range(N3)]
                else:
                    N1=int(np.ceil(np.log(length*(rmax-1)/deltamin+1.)/np.log(rmax)))
                    while(1):
                        temp3=deltamin*(np.power(r,N1)-1)/(r-1)
                        if (np.abs(temp3/length-1)<self.sensitivity):
                            break
                        elif (temp3>length):
                            r=r*cc
                        elif (temp3<length):
                            r=r*(2.0-cc)
                    error = length-temp3
                    grids = grids + [deltamin*np.power(r,i) for i in range(N1)]
                    grids[-1] += error
            if (shape==1):
                grids.reverse()
        return grids

    def SetUnit(self,s):
        self.unit=s

    def get_nof_sub(self, eksen):
        if (eksen==0):
            return self.sizex-1
        elif (eksen==1):
            return self.sizey-1
        elif (eksen==2):
            return self.sizez-1
        return 0

if __name__ == "__main__":
    # krix=[0,1.0,1.05,2.0]
    # aa=CartesianGrid()
    # aa.findsubregions(krix,[],[])
    # aa.nu_grid(0.1,0.1,0.1)
    # aa.optimizegrids()
    # import matplotlib.pyplot as plt
    # plt.plot(aa.dx)
    # plt.show()
    cg = CartesianGrid()
    import time
    t1 = time.time()
    cg.nugrid(1.0,2.0,1.2,20,0, 0,cc=0.999)
    t2 = time.time()
    cg.nugrid(1.0,2.0,1.2,20,0, 0,cc=0.99)
    t3 = time.time()
    cg.nugrid(1.0,2.0,1.2,20,0, 0,cc=0.9)
    t4 = time.time()
    cg.nugrid(1.0,2.0,1.2,20,0, 0,cc=0.5)
    t5 = time.time()
    print("t2-t1= "+str(t2-t1))
    print("t3-t2= "+str(t3-t2))
    print("t4-t3= "+str(t4-t3))
    print("t5-t4= "+str(t5-t4))
    pass
